/*
 * LineRefParser.java       Jan 13, 2007
 *
 * Copyright (c) 2007, Brian Mulanda, all rights reserved.
 */

package edu.ksu.cis.ndpc;

import java.util.ArrayList;

/**
 * Parses a string containing a reference to a line or set of lines enclosed in
 * a natural deduction proof box, generating a line reference. Multiple
 * references in the string are separated by commas.
 * 
 * @author Brian Mulanda (<a
 *         href="mailto:bmulanda@ksu.edu">bmulanda@ksu.edu</a>)
 * 
 * @see LineReference
 */

public class LineRefParser {
	private static final int T_INVALID = -1;
	private static final int T_INTEGER = 0;
	private static final int T_HYPHEN = 1;
	private static final int T_COMMA = 2;
	private static final int T_NULL = 3;
	private boolean[][] syntaxTable;

	public LineRefParser() {
		initializeSyntaxTable();
	}

	public LineReference[] parse(String lineref) throws LineRefParserException {
		ArrayList linerefs = null;
		LineReference lineRef = null;
		Tokenizer tokenizer = null;
		Token token = null;
		Token prevToken = null;
		int type = T_NULL;
		int prevType = T_NULL;
		boolean done = false;
		boolean range = false;

		if (lineref != null && lineref.trim().length() > 0) {
			linerefs = new ArrayList();
			tokenizer = new Tokenizer(lineref.trim(), ",-");
			while (!done) {
				if (tokenizer.nextToken()) {
					token = tokenizer.getToken();
					type = identify(token.getValue());
				} else {
					type = T_NULL;
				}

				if (type == T_INVALID) {
					throw new LineRefParserException(
							"Invalid character(s) in line reference: " + token);
				} else {
					if (syntaxTable[prevType][type]) {
						switch (type) {
						case T_COMMA:
							break;
						case T_HYPHEN:
							break;
						case T_INTEGER:
							if (Integer.parseInt(token.getValue()) > 0) {
								if (range) {
									lineRef = new Range(Integer
											.parseInt(prevToken.getValue()),
											Integer.parseInt(token.getValue()));
									linerefs.add(lineRef);
									range = false;
								} else {
									if (tokenizer.nextToken()) {
										if (identify(tokenizer.getToken()
												.getValue()) == T_HYPHEN) {
											range = true;
											prevToken = token;
										}
										tokenizer.pushBack();
									}
									if (!range) {
										lineRef = new Line(Integer
												.parseInt(token.getValue()));
										linerefs.add(lineRef);
									}
								}
							} else {
								throw new LineRefParserException(
										"Invalid number in line reference: "
												+ token);
							}
							break;
						}
					} else {
						throw new LineRefParserException(
								"Unexpected character(s) in line reference: "
										+ token);
					}
					prevType = type;
					done = (type == T_NULL);
				}
			}
		}

		if (linerefs == null) {
			return null;
		} else {
			return (LineReference[]) linerefs.toArray(new LineReference[0]);
		}
	}

	private int identify(String str) {
		int type = T_INVALID;

		if (str.equals(",")) {
			type = T_COMMA;
		} else if (str.equals("-")) {
			type = T_HYPHEN;
		} else if (isInteger(str)) {
			type = T_INTEGER;
		}
		return type;
	}

	private void initializeSyntaxTable() {
		syntaxTable = new boolean[][] {

		/*
		 * n - , EOF / n
		 */{ false, true, true, true },
		/* - */{ true, false, false, false },
		/* , */{ true, false, false, false },
		/* EOF */{ true, false, false, false },

		};
	}

	private boolean isInteger(String str) {
		boolean ok = true;
		try {
			Integer.parseInt(str);
		} catch (NumberFormatException nfe) {
			ok = false;
		}
		return ok;
	}
}
