package edu.ksu.cis.ndpc;

/**
 * @author Brian Mulanda
 *
 */

import java.util.Vector;

import edu.ksu.cis.ndpc.gui.RawProof;
import edu.ksu.cis.ndpc.gui.RawStatement;
import edu.ksu.cis.ndpc.rules.*;

public class ProofChecker {
	private static LineRefParser lineParser = new LineRefParser();

	public void check(RawProof rawProof, LangParser lParser)
			throws ProofCheckerException {
		checkProof(buildProof(rawProof), lParser);
	}

	public Proof buildProof(RawProof rproof) throws ProofCheckerException {
		Proof proof = new Proof();
		ProofStatement statement = null;
		RawStatement rstat = null;

		int statcount = 0;
		int laststat = 0;
		int lineId = 1;
		for (int i = 0; i < rproof.size(); ++i) {
			rstat = rproof.getStatement(i);
			int boxesBegin = rproof.boxesAtStat(i)[0];
			int boxesEnd = rproof.boxesAtStat(i)[1];

			String var = rstat.getVariable();
			String form = rstat.getFormula();
			String rule = rstat.getRule();
			String lineref = rstat.getLineref();

			statement = new ProofStatement(lineId);
			if (form != null) {
				statcount++;
				laststat = lineId;
			}

			ProofVariable v = (var != null) ? new ProofVariable(var) : null;
			ProofFormula f = (form != null) ? new ProofFormula(form) : null;
			ProofRule r = (rule != null) ? new ProofRule(rule) : null;
			ProofLineReference l = (lineref != null) ? new ProofLineReference(
					lineref) : null;

			statement.variable = v;
			statement.formula = f;
			statement.rule = r;
			statement.lineref = l;

			for (int j = 0; j < boxesBegin; ++j) {
				proof.openBox();
			}

			proof.addStatement(statement);

			for (int j = 0; j < boxesEnd; ++j) {
				proof.closeBox();
			}
			lineId++;
		}

		proof.setActualStatCount(statcount);
		proof.setLastStatNo(laststat);
		return proof;
	}

	/*
	 * Check proof line by line.
	 * 
	 * For each formula: 1. Parse the formula to generate a parse tree 2. Parse
	 * the reason/justification to identify the rule 3. Parse the
	 * reason/justification to ensure existence of the parameters of the rule.
	 * 4. Parse the reason/justification to ensure validity of references to
	 * other statements in the proof. 5. Prepare the parameters required by the
	 * rule 6. Check the rule
	 */

	private void checkProof(Proof proof, LangParser lParser)
			throws ProofCheckerException {
		if (proof == null || proof.getActualStatCount() <= 0) {
			throw new ProofCheckerException("Nothing to check");
		} else if (proof.statIsInBox(proof.getLastStatNo())) {
			throw new ProofCheckerException("This is not a valid proof "
					+ "because the last statement (line no. "
					+ proof.getLastStatNo() + ") " + "is inside a box.");
		} else {
			int line = 1;
			try {
				ProofStatement statement = null;

				for (line = 1; line <= proof.getStatementCount(); ++line) {
					statement = proof.getStatement(line);
					checkStatement(proof, statement, lParser);
				}
			} catch (ProofCheckerException pfce) {
				// pfce.printStackTrace();
				throw new ProofCheckerException("Error on line: " + line + "\n"
						+ pfce.getMessage());
			}
		}
	}

	private void checkVariable(Proof proof, ProofStatement statement,
			LangParser lParser) throws ProofCheckerException {

		Symbol variable = statement.getVariable();
		variable.type = lParser.getSymbols().identify(variable.getValue());

		int lineno = statement.getLineNo();
		if (variable.getType() == LangSymbols.S_CONST) {
			throw new ProofCheckerException("'" + variable
					+ "' is a constant; you cannot use "
					+ "a constant in the place of a fresh variable.");
		} else if (!proof.statIsFirstInBox(lineno)) {
			throw new ProofCheckerException(
					"Fresh variable '"
							+ variable
							+ "' cannot be introduced here; "
							+ "A fresh variable can be introduced only on the first line in a box.");
		} else if (proof.isUsedVariable(variable)) {
			throw new ProofCheckerException(
					"Fresh variable '"
							+ variable
							+ "' "
							+ "has already been introduced or used earlier in the proof.");
		}
		proof.addFreshVariable(variable, lineno);
	}

	private void checkFormula(Proof proof, ProofStatement statement,
			LangParser lParser) throws ProofCheckerException {
		ProofFormula formula = statement.getFormula();
		try {
			lParser.parse(formula);
			Symbol variable = null;
			if (formula.variables != null) {
				for (int i = 0; i < formula.variables.size(); ++i) {
					variable = (Symbol) formula.variables.get(i);
					if (proof.isFreshVariable(variable)
							&& !proof
									.isInScope(variable, statement.getLineNo())) {
						throw new ProofCheckerException(
								"'"
										+ variable
										+ "' is a "
										+ "fresh variable; it cannot be used here. A fresh variable "
										+ "should not appear anywhere outside its box");
					}
					proof.addVariable(variable);
				}
			}
		} catch (Exception e) {
			// e.printStackTrace();
			throw new ProofCheckerException(e.getMessage());
		}
	}

	private void checkRule(Proof proof, ProofStatement statement)
			throws ProofCheckerException {
		String rule = statement.getRule().rule;
		int ruleType = Rule.identify(rule);
		if (ruleType == Rule.INV) {
			throw new ProofCheckerException("Rule " + rule
					+ " is not a valid rule");
		} else if (ruleType == Rule.ASSUMPTION
				&& !proof.statIsFirstInBox(statement.getLineNo())) {
			throw new ProofCheckerException(
					"This formula is an assumption; an assumption "
							+ "must be the first formula inside a box");
		}

		statement.getRule().type = ruleType;
	}

	private void checkLineref(Proof proof, ProofStatement statement)
			throws ProofCheckerException {
		try {
			LineReference[] linerefs = lineParser
					.parse(statement.getLineref().lineref);
			LineReference lineref = null;
			for (int i = 0; i < linerefs.length; ++i) {
				lineref = linerefs[i];
				if (lineref instanceof Range) {
					if (!proof.rangeIsBox(lineref)) {
						throw new ProofCheckerException(lineref
								+ " does not reference a valid box.");
					}
					// System.out.println(lineref + " is box");
					if (!proof.canReferenceBox(statement.getLineNo(), lineref)) {
						throw new RuleException(
								"This formula cannot reference box: " + lineref
										+ ".");
					}
				} else {
					if (proof.getStatement(lineref.getBegin()).getFormula() == null) {
						throw new RuleException(
								"This formula is referencing a line, "
										+ lineref.getBegin()
										+ ", that does not contain a formula.");
					} else if (!proof.canReferenceLine(statement.getLineNo(),
							lineref.getBegin())) {
						throw new RuleException(
								"This formula cannot reference line: "
										+ lineref + ".");
					}
				}
			}
			statement.getLineref().linerefs = linerefs;
		} catch (ProofCheckerException pcke) {
			throw pcke;
		} catch (Exception e) {
			// e.printStackTrace();
			throw new ProofCheckerException(e.getMessage());
		}
	}

	private void checkStatement(Proof proof, ProofStatement statement,
			LangParser lParser) throws ProofCheckerException {
		if (statement.getFormula() != null && statement.getRule() == null) {
			throw new ProofCheckerException("Missing rule for formula");
		} else if (statement.getFormula() == null
				&& (statement.getRule() != null || statement.getLineref() != null)) {

			throw new ProofCheckerException("Missing formula");
		}

		if (statement.getVariable() != null) {
			checkVariable(proof, statement, lParser);
		}

		if (statement.getFormula() != null) {
			checkFormula(proof, statement, lParser);
		}

		if (statement.getRule() != null) {
			checkRule(proof, statement);
		}

		if (statement.getLineref() != null) {
			checkLineref(proof, statement);
		}

		if (statement.getRule() != null) {
			Rule rule = Rule.getRule(statement.getRule().type);
			try {
				RuleParameter[] params = null;
				if (statement.getLineref() != null
						&& statement.getLineref().linerefs != null) {
					LineReference[] linerefs = statement.getLineref().linerefs;
					params = new RuleParameter[linerefs.length + 1];
					LineReference lineref = null;
					for (int i = 0; i < linerefs.length; ++i) {
						lineref = linerefs[i];
						if (lineref instanceof Range) {
							params[i] = (RuleParameter) proof
									.getBoxedStatements((Range) lineref);
						} else {
							params[i] = (RuleParameter) proof
									.getStatement(lineref.getBegin());
						}
					}
				} else {
					params = new RuleParameter[1];
				}
				params[params.length - 1] = (RuleParameter) statement;
				rule.check(params);
			} catch (RuleException rle) {
				throw new ProofCheckerException(
						"Incorrect application of Rule: "
								+ Rule.getDescription(statement.getRule().type)
								+ "\n" + rle.getMessage());
			} catch (Exception e) {
				// e.printStackTrace();
				throw new ProofCheckerException(e.getMessage());
			}
		}
	}

	public static ParseTree substitute(ParseTree formula, ParseTree termF,
			ParseTree termR) {
		if (formula != null && termF != null && termR != null
				&& !termF.equals(termR)) {
			int type = formula.getRootNode().getType();
			boolean subchild = true;

			if (type == LangSymbols.S_FORALL || type == LangSymbols.S_EXISTS) {
				ParseTree var = new ParseTree(new Symbol(formula.getRootNode()
						.getValue().substring(1), LangSymbols.S_VAR));
				if (termF.containsLeaf(var) || termR.containsLeaf(var)) {
					subchild = false;
				}
			} else if (PredLSymbols.isTerm(type) && formula.equals(termF)) {
				formula.replaceWith(termR);
				subchild = false;
			} else if (formula.isLeaf()) {
				subchild = false;
			}
			if (subchild) {
				int children = formula.getSubtreeCount();
				for (int i = 0; i < children; ++i) {
					substitute(formula.getSubtree(i), termF, termR);
				}
			}
		}
		return formula;
	}

	public static void checkSubstitution(ParseTree fold, ParseTree fnew,
			ParseTree termT, ParseTree termX, boolean partial)
			throws ProofCheckerException {
		checkSubstitution(fold, fnew, termT, termX, null, partial);
	}

	public static void checkSubstitution(ParseTree fold, ParseTree fnew,
			ParseTree termT, ParseTree termX, ParseTree scope)
			throws ProofCheckerException {
		checkSubstitution(fold, fnew, termT, termX, scope, false);
	}

	public static ParseTree findTermT(ParseTree treeX, ParseTree treeT,
			ParseTree termX) {
		ParseTree tree = null;
		if (PredLSymbols.isTerm(treeX.getRootNode().getType())) {
			if (!treeX.getRootNode().equals(treeT.getRootNode())
					&& treeX.equals(termX)) {
				tree = treeT;
			}
		}
		if (tree == null && !treeX.isLeaf() && !treeT.isLeaf()) {
			if (!((treeX.getRootNode().getType() == LangSymbols.S_FORALL || treeX
					.getRootNode().getType() == LangSymbols.S_EXISTS) && treeX
					.getRootNode().endsWith(termX.getRootNode()))) {

				if (!treeX.getRootNode().equals(termX.getRootNode())) {
					ParseTree[] x_subtrees = treeX.getSubtrees();
					ParseTree[] t_subtrees = treeT.getSubtrees();

					for (int i = 0; tree == null && i < x_subtrees.length; ++i) {
						tree = findTermT(x_subtrees[i], t_subtrees[i], termX);
					}
				}
			}
		}
		return tree;
	}

	public static void checkSubstitution(ParseTree fold, ParseTree fnew,
			ParseTree termT, ParseTree termX, ParseTree scope, boolean partial)
			throws ProofCheckerException {

		if (fold != null && fnew != null) {
			if (!fold.getRootNode().equals(fnew.getRootNode())) {
				if (PredLSymbols.isTerm(fold.getRootNode().getType())
						&& fold.equals(termX) && fnew.equals(termT)) {
					if (!isFreeForIn(termT, fold, scope)) {
						throw new ProofCheckerException("Term: '"
								+ toString(fold) + "' is not free for "
								+ "term '" + toString(termT) + "'");
					}
				} else {
					throw new ProofCheckerException("'" + toString(fnew)
							+ "' should remain as " + "'" + toString(fold)
							+ "'");
				}
			} else {
				if (!partial
						&& PredLSymbols.isTerm(fold.getRootNode().getType())
						&& fold.equals(termX) && !termT.equals(termX)
						&& isFreeForIn(termT, fold, scope)) {

					throw new ProofCheckerException("All free occurrences of '"
							+ termX + "' should be replaced " + "by '" + termT
							+ "'.There is at least one free occurrence of '"
							+ termX + "' " + "that has not been replaced");
				} else {
					ParseTree[] fold_subtrees = fold.getSubtrees();
					ParseTree[] fnew_subtrees = fnew.getSubtrees();

					if (fold_subtrees != null && fnew_subtrees != null) {
						int count = 0;
						if (fold_subtrees.length < fnew_subtrees.length) {
							count = fold_subtrees.length;
						} else {
							count = fnew_subtrees.length;
						}
						for (int i = 0; i < count; ++i) {
							checkSubstitution(fold_subtrees[i],
									fnew_subtrees[i], termT, termX, scope,
									partial);
						}
					}
				}
			}
		}
	}

	public static boolean isFree(ParseTree termX) {
		return isFreeForIn(null, termX, null);
	}

	public static boolean isFree(ParseTree termX, ParseTree scopetop) {
		return isFreeForIn(null, termX, scopetop);
	}

	public static boolean isFreeFor(ParseTree termT, ParseTree termX) {
		return isFreeForIn(termT, termX, null);
	}

	public static boolean isFreeForIn(ParseTree termT, ParseTree termX,
			ParseTree scopetop) {
		boolean free = true;
		ParseTree currentTree = termX.getParent();
		while (free && currentTree != null
				&& (scopetop == null || currentTree != scopetop)) {
			Symbol root = currentTree.getRootNode();
			if (root.getType() == PredLSymbols.S_FORALL
					|| root.getType() == PredLSymbols.S_EXISTS) {
				ParseTree leaf = null;
				leaf = new ParseTree(new Symbol(root.getValue().substring(1),
						PredLSymbols.S_VAR));
				if (termX.containsLeaf(leaf)
						|| (termT != null && termT.containsLeaf(leaf))) {
					free = false;
				}
			}

			if (free) {
				currentTree = currentTree.getParent();
			}
		}
		return free;
	}

	public static boolean isFreeIn(ParseTree termT, ParseTree formula) {
		boolean free = true;
		if (formula != null && termT != null) {
			int type = formula.getRootNode().getType();

			if (type == LangSymbols.S_FORALL || type == LangSymbols.S_EXISTS) {
				Symbol var = new Symbol(formula.getRootNode().getValue()
						.substring(1), LangSymbols.S_VAR);
				if (var.equals(termT.getRootNode())) {
					free = false;
				}
			} else if (!formula.isLeaf()) {
				int children = formula.getSubtreeCount();
				for (int i = 0; i < children && free; ++i) {
					free = isFreeIn(termT, formula.getSubtree(i));
				}
			}
		}
		return free;
	}

	public static int countFreeIn(ParseTree termX, ParseTree formula) {
		return countFreeForIn(null, termX, formula);
	}

	public static int countFreeForIn(ParseTree termT, ParseTree termX,
			ParseTree formula) {
		int count = 0;
		if (formula != null) {
			int type = formula.getRootNode().getType();
			boolean subchild = true;

			if (type == LangSymbols.S_FORALL || type == LangSymbols.S_EXISTS) {
				ParseTree var = new ParseTree(new Symbol(formula.getRootNode()
						.getValue().substring(1), LangSymbols.S_VAR));
				if ((termT != null && termT.containsLeaf(var))
						|| termX.containsLeaf(var)) {
					subchild = false;
				}
			} else if (PredLSymbols.isTerm(type) && formula.equals(termX)) {
				count++;
				subchild = false;
			} else if (formula.isLeaf()) {
				subchild = false;
			}
			if (subchild) {
				int children = formula.getSubtreeCount();
				for (int i = 0; i < children; ++i) {
					count += countFreeForIn(termT, termX, formula.getSubtree(i));
				}
			}
		}
		return count;
	}

	public static ParseTree getNode(ParseTree tree, int[] path) {
		if (tree != null && path != null && path.length > 0) {
			for (int i = 0; i < path.length; ++i) {
				tree = tree.getSubtree(path[i]);
				// System.out.println(" subtree("+i+"):"+tree);
				if (tree == null) {
					i = path.length;
				}
			}
		}
		return tree;
	}

	public static Object[] getTerms(ParseTree tree) {
		Vector terms = new Vector();

		getAllTerms(terms, tree);
		return terms.toArray();
	}

	private static void getAllTerms(Vector terms, ParseTree tree) {
		if (tree != null) {
			if (PredLSymbols.isTerm(tree.getRootNode().getType())) {
				terms.add(tree);
			} else {
				ParseTree[] subtrees = tree.getSubtrees();
				if (subtrees != null) {
					for (int i = 0; i < subtrees.length; ++i) {
						getAllTerms(terms, subtrees[i]);
					}
				}
			}
		}
	}

	/* For debugging */
	public static String getPathString(int[] path) {
		StringBuffer pathStr = new StringBuffer();
		for (int j = 0; j < path.length; ++j) {
			pathStr.append(path[j] + "->");
		}
		return pathStr.toString();
	}

	public static boolean equalSyntax(ParseTree treeX, ParseTree treeT) {
		return getStruct(treeX).equalsIgnoreCase(getStruct(treeT));
	}

	public static String getStruct(ParseTree tree) {
		StringBuffer struct = new StringBuffer(0);

		if (tree != null) {
			ParseTree[] subtrees = tree.getSubtrees();
			if (subtrees != null) {
				for (int i = 0; i < subtrees.length; ++i) {
					struct.append(getStruct(subtrees[i]));
				}
			}
			if (!PredLSymbols.isTerm(tree.getRootNode().getType())) {
				struct.append(tree.getRootNode());
			}
		}
		return struct.toString();
	}

	public static ParseTree[] getSyntax(ParseTree tree) {
		Vector syntax = new Vector();

		getTreeSyntax(syntax, tree);

		ParseTree[] syntaxArray = new ParseTree[syntax.size()];
		for (int i = 0; i < syntax.size(); ++i) {
			syntaxArray[i] = (ParseTree) syntax.get(i);
		}
		return syntaxArray;
	}

	public static String toString(ParseTree tree) {
		String formula = "";
		if (tree != null) {
			formula = tree.getRootNode().toString();
			if (tree.isLeaf()) {

			} else {
				ParseTree[] subtrees = tree.getSubtrees();
				if (subtrees.length == 1) {
					formula = tree.getRootNode().toString();
					formula = formula + "(" + toString(subtrees[0]) + ")";
				} else if (subtrees.length == 2) {
					if (tree.getRootNode().getType() == LangSymbols.S_PRED
							|| tree.getRootNode().getType() == LangSymbols.S_FUNC) {

						formula = tree.getRootNode().toString() + "(";
						formula = formula + toString(subtrees[0]) + ",";
						formula = formula + toString(subtrees[1]);
						formula = formula + ")";

					} else {
						formula = "(" + toString(subtrees[0]) + ")";
						formula = tree.getRootNode().toString();
						formula = "(" + toString(subtrees[1]) + ")";

					}
				} else {
					formula = tree.getRootNode().toString() + "(";
					for (int i = 0; i < subtrees.length; ++i) {
						formula = formula + toString(subtrees[i]) + ",";
					}
					formula = formula.substring(0, formula.length() - 1);
					formula = formula + ")";
				}
			}
		}
		return formula;
	}

	/* tree must not be null */
	public static ParseTree[] getNodes(ParseTree tree) {
		return getNodes(tree, null, false);
	}

	public static ParseTree[] getNodes(ParseTree tree, ParseTree sample) {
		return getNodes(tree, sample, false);
	}

	public static ParseTree[] getNodes(ParseTree tree, boolean childTerms) {
		return getNodes(tree, null, childTerms);
	}

	public static ParseTree[] getNodes(ParseTree tree, ParseTree sample,
			boolean childTerms) {
		Vector nodes = listNodes(tree, sample, childTerms);
		ParseTree[] nodesArray = new ParseTree[nodes.size()];
		nodes.toArray(nodesArray);
		return nodesArray;
	}

	private static Vector listNodes(ParseTree tree, ParseTree sample,
			boolean childTerms) {
		Vector nodes = new Vector();
		if (tree != null) {
			nodes.add(tree);
			Symbol root = tree.getRootNode();
			if (childTerms || !PredLSymbols.isTerm(root.getType())) {
				if (!tree.isLeaf() && tree.equals(sample)) {
					ParseTree[] subtrees = tree.getSubtrees();
					for (int i = 0; i < subtrees.length; ++i) {
						ParseTree child = (ParseTree) subtrees[i];
						nodes.addAll(listNodes(child, sample, childTerms));
					}
				}
			}
		}
		return nodes;
	}

	public static ParseTree[] getLeaves(ParseTree tree) {
		return getLeaves(tree, null);
	}

	public static ParseTree[] getLeaves(ParseTree tree, ParseTree sample) {
		Vector leaves = listLeaves(tree, sample);
		ParseTree[] leavesArray = new ParseTree[leaves.size()];
		leaves.toArray(leavesArray);
		return leavesArray;
	}

	private static Vector listLeaves(ParseTree tree, ParseTree sample) {
		Vector leaves = new Vector();
		if (tree != null) {
			if (tree.isLeaf()) {
				if (sample == null || tree.equals(sample)) {
					leaves.add(tree);
				}
			} else {
				ParseTree[] subtrees = tree.getSubtrees();
				for (int i = 0; i < subtrees.length; ++i) {
					ParseTree child = (ParseTree) subtrees[i];
					leaves.addAll(listLeaves(child, sample));
				}
			}
		}
		return leaves;
	}

	private static void getTreeSyntax(Vector syntax, ParseTree tree) {
		if (tree != null) {
			syntax.add(tree);
			if (!PredLSymbols.isTerm(tree.getRootNode().getType())) {
				ParseTree[] subtrees = tree.getSubtrees();
				if (subtrees != null) {
					for (int i = 0; i < subtrees.length; ++i) {
						getTreeSyntax(syntax, subtrees[i]);
					}
				}
			}
		}
	}
}
